/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Log: parse_tag.c,v $
 * Revision 1.27  2000/05/09 14:35:15  gellert
 * after a illegal attribut go on with the next attribut (after the
 * next whitespace).
 *
 * Revision 1.26  2000/05/08 13:13:16  glauschwuffel
 * minor fixes for attribute processing
 *
 * Revision 1.25  2000/05/07 16:13:42  glauschwuffel
 * made functions local. now we process documents this way: parse each item bit by bit. we do not know the end of a tag anymore when entering parse_tag. this is the first step to be robust against unmatched quotes.
 *
 * Revision 1.24  2000/04/24 19:32:54  gellert
 * cleaned up the return values to get rid of integer-pointer warnings.
 *
 * Revision 1.23  2000/04/18 15:11:36  gellert
 * unallowed_href returned -1 if the attribute had no value, so
 * attributes without values would not be allowed. fixed.
 *
 * Revision 1.22  2000/04/16 15:38:58  gellert
 * fixed two bugs:
 * a) check_for_script was not called. and now it checks all the
 *    ampersands for a following brace, not just the first one.
 * b) we have to call proto_allowed not only for HREF, but also
 *    for the SRC attribute.
 *
 * Revision 1.21  2000/03/07 15:06:54  goldbach
 * added GPL info
 *
 * Revision 1.20  2000/02/21 14:11:52  gellert
 * And we shouldn't call logmsg with NULL-pointers... ;-)
 *
 * Revision 1.19  2000/02/21 13:57:50  gellert
 * In "unallowed_href": inserted a test, if the given value is
 * a NULL-pointer. Otherwise httpf segfaults at <A HREF>...
 *
 * Revision 1.18  2000/02/21 13:42:47  gellert
 * In the valeu of a HREF we now search not only for ":", but
 * also for &#58, to prevent the values "javascript&#58"
 *
 * Revision 1.17  1999/11/26 10:47:08  goldbach
 * New actions introduced: replskip and replendskip skip sections of documents, i.e. data won\'t be sent to the client.
 *
 * Revision 1.16  1999/11/12 12:16:20  gellert
 * Fixed:	If an attribute was logged only (and not replaced or
 * 	allowed), the attributes were processed afterwards
 * 	(so the "<" and the tag were ommitted, but the attributes
 * 	and the ">" were copied to the outbuf).
 *
 * Revision 1.15  1999/11/11 15:58:02  gellert
 * Removed declarations of some nonexistent functions.
 * Some new logmessages inserted.
 *
 * Revision 1.14  1999/08/09 12:09:32  goldbach
 * Changed unallowed_href to accept hrefs without *any* protocols at all :)
 * Bug: do_attr aborted after processing an attribute without a value. This
 * buggy behaviour must have been around since the removal of the high-level
 * strncpy() for everything between the angles... tables look fine now :)
 *
 * Revision 1.13  1999/07/28 14:43:05  gellert
 * moved some functions to misc.c
 *
 * Revision 1.12  1999/07/22 15:35:04  goldbach
 * yet again cleanup for logmsg()
 *
 * Revision 1.11  1999/07/22 14:30:52  gellert
 * some improvements of the determination of the tag end,
 * now nested quotes etc are supported.
 * changed config option replace to replabort/replcont.
 *
 * Revision 1.10  1999/07/15 17:01:22  gellert
 * added the check of href tags for allowed protocols.
 *
 * Revision 1.9  1999/07/14 09:50:06  goldbach
 * removed last strncpy.
 *
 * Revision 1.8  1999/06/30 14:10:15  goldbach
 * removed subtle bug in process attr
 *
 * Revision 1.7  1999/06/29 11:56:41  gellert
 * parsing changed: now there are 0-bytes written into the
 * input buffer to terminate the strings. so they need not
 * to be copied into some extra buffer...
 *
 * Revision 1.6  1999/06/23 11:14:40  gellert
 * deleted some never reached line
 *
 * Revision 1.5  1999/06/23 10:45:20  gellert
 * better checking of the tag,
 * bug fixes (segfaults).
 *
 *
 * $Id: parse_tag.c,v 1.27 2000/05/09 14:35:15 gellert Exp $
 */

#include <syslog.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <ctype.h>
#include "mythreads.h"
#include "taghash.h"
#include "cfglists.h"
#include "misc.h"
#include "parse_tag.h"

int parse_tag(char*, TCT*);

static int process_tag(char*, TCT*);
static int check_tag(char*, TCT*);
static int do_tag(char*, TCT*);
char *skip_until_end_of_tag(char*);
static char *find_tagstart(char*, TCT*);
static char *find_tagend(char*);
static char *find_attrstart(char*, TCT*);
static char *find_attrend(char*);
static char *find_valstart(char*,TCT*);
static char *find_valend(char*, TCT*); 
static char *do_attr(char*, TCT*);
static char *process_attr(char*, char*, TCT*);
static int check_for_script(char*);

static int unallowed_href(char*, char*);

static int copy_doctype(char*, TCT*);

/*
 * unallowed_href
 *
 * Checks if a hyperlink is allowed or not. It is allowed if it's
 * a reference without protocol or a reference with a
 * configured protocol.
 * Note:
 *   Currently this works only with the colon character ':' as protocol
 *   delimiter. We search &#xx, too. At least &#58 (that's ":")
 *   If the protocol name contains any &#xx, proto_allowed will not
 *   find them (and thus not allow them)
 */
static int unallowed_href(char *attr, char *value) {
    char *temp, *temp1, *temp2;

    /* do we have a value? */
    if (value == NULL) {
      return(0);
    }

    /* do we have a link or a src? */
    if ((strncasecmp("href", attr, 5) != 0) &&
        (strncasecmp("src", attr, 4) != 0)) {
	return(0);
    }
    
    /* look for a colon in the attribute's value */
    temp1 = strnchr(value, ':', strlen(value));
    temp2 = strstr(value, "&#58");

    if ((temp1 == NULL) && (temp2 == NULL)) {
	/* there is no colon */
	/* we don't have a protocol, so the reference is okay */
	return(0);
    }

    temp = value;
    while ((*temp== ' ') || (*temp== '\"') || (*temp=='\'')) {
	temp++;
    }
    
    
    if (proto_allowed(temp)) {
	return(-1);
    } else {
	return(0);
    }
}



/*
 * find_tagstart
 *
 * Find start of tag.
 * See find_attrstart, the only difference is that we allow
 * a '/' at the start. 
 */
static char *find_tagstart(char *buf, TCT *tct) {
  char *tag_start;  

  /* skip whitespace */
  tag_start = skip_leading_whitespace(buf);
    
  if(legal_constituent(tag_start[0]) || tag_start[0]=='/') {
	/* set recovery point */
	tct->last_valid_tag = tag_start;
	return tag_start;
  } else
	return NULL;
}


/*
 * find_tagend
 *
 * Find end of tag.
 * Takes pointer to beginning of tag.
 * Result points right behind the tag or NULL if
 * tag contains illegal characters.
 */
static char *find_tagend(char *tag_start) {
  int i=0; char *eot;
 
  /* Find end of legal characters. */
  while(legal_constituent(tag_start[i])) {
	i+=1;
  }

  eot=tag_start+i;

  /* tag ends with white space or '>', everything else indicates failure */
  if (legal_whitespace(eot[0]) || eot[0]=='>')
	return eot;
  else
	return NULL;
}

/*
 * find_attrstart
 *
 * Find start of attribute.
 * Takes pointer to buffer that hopefully contains attributes.
 * Leading whitespace is skipped.
 * Returns pointer to first character of attribute or NULL if
 * attribute starts with illegal characters.
 */
static char *find_attrstart(char *buf, TCT *tct) {
  char *attr_start;  

  /* skip whitespace */
  attr_start = skip_leading_whitespace(buf);
  
  if(legal_constituent(attr_start[0])) {
	/* set recovery point */
	tct->last_valid_attr = attr_start;
	return attr_start;
  } else
	return NULL;
}

/*
 * find_attrend
 *
 * Find end of attribute.
 * Takes pointer to beginning of attribute. 
 * Result points right behind last character of attribute or NULL if
 * attribute contains illegal characters.
 */
static char *find_attrend(char *attr_start) {
  int i=0;
    
  /* Find end of legal characters. */
  while(legal_constituent(attr_start[i])) {
	i+=1;
  }

  /*
   * an attribute name ends with
   *  - '='
   *      we have a value
   *  - whitespace (there might be a value)
   * - '>'
   *     This is the last attribute
   */
  if (attr_start[i]=='=' ||
      attr_start[i]=='>' ||
	  legal_whitespace(attr_start[i]))
	return attr_start+i;
  else {
	return NULL;
	}
}

/*
 * find_valstart
 *
 * Find start of (optional) value for an attribute.
 * BUF points to the start of the current attribute.
 *
 * Notes: we allow a hash ('#') at the start of a value
 *   if it is not followed by an ampersand ('&') since this
 *   would allow entities...
 *
 * Returns:
 * -1 - illegal value
 *  0 - no value
 *  other - start of value.
 */

static char *find_valstart(char *attr_end, TCT *tct) {
  char *cont;
  
  /* we should have a '=' or a whitespace now */
  if (legal_whitespace(attr_end[0]) != 0) {
	cont=skip_leading_whitespace(attr_end);
  } else {
	cont=attr_end;
  }

  /* no value if we are at end of string */
  if (cont[0]==0 || cont[0]=='>') return NULL;
  
  /*
   * now the only valid continuations are:
   *   '=' -- we have a value
   *   legal_constituent -- new attribute
   * Other values are an error.
   */
  if (cont[0]=='=') {
	/* we have a value */
	cont = skip_leading_whitespace(cont+1);
	if (legal_constituent(cont[0]) != 0 ||
		(cont[0]=='#' && cont[1]!='&') || /* no entities ;) */
		a_quote(cont)) {
	  /* valstart is legal, set recovery point */
	  tct->last_valid_value = cont;
	  return cont;
	} else {
	  /* valstart is illegal */
	  return (char *)-1;
	}
  }
  
  /* we have no value or a malformed input */
  if (legal_constituent(cont[0]) != 0)
	return NULL; /* no value (new attr) */
  else
	return (char *)-1; /* illegal continuation */
}

/*
 * find_valend
 *
 * Find end of value of current attribute.
 * Inputs:
 *  valstart - result of find_valstart
 * Returns:
 *  address of end of value.
 */
static char *find_valend(char *valstart, TCT *tct) {
  char c;

  if(valstart==NULL) return NULL;

  /*
   * First we check for a heading quote.
   * These comparisons are based on the HTML 4.0
   * specs (3.2.2 Attributes, p. 25).
   * This function assumes this:
   *   - quotes should be nested correctly
   *   - &#39 is treated as a single quote and
   *     &#34 and &quot; as double quote. Thus: the string
   *       "bla&quot;
   *     looks awkward, however, it is correct.
   */
  c = a_quote(valstart);
  if (c) {
	/* value starts with a quote */
	return find_end(valstart+1, c, tct);
  } else {
	/* no quote */
	return find_unquoted_end(valstart+1);
  }
}

/*
 * do_attr
 *
 * Try to get and handle the next attribute in a string.
 * Returns:
 *  0  -- abort parsing this string. 
 *        The attr is not allowed so skip the rest.
 *  FAILURE -- there was an error (illegal characters in attribute or value).
 *  other --  the string right behind the attribute's value or the string
 *            right behind the attribute (if there was no value)
 *
 * Note: we modify the input buffer. The attribute and value string
 *       are terminated with a null character.
 */
static char *do_attr(char *buf, TCT *tct) {
  char *attr_start, *attr_end, *val_start=NULL, *val_end;
  char replaced;

  /* find attribute */
  attr_start = find_attrstart(buf,tct);
  if (attr_start==NULL)  /* illegal chars? */
	return (char *)FAILURE;
    
  attr_end = find_attrend(attr_start);
  if (attr_end==NULL) { /* illegal chars? */
	/* jump to next attribute */
	tct->inoffset=skip_to_next_whitespace(attr_start)-tct->inbuf+1;
	logmsg(LOG_DEBUG, "[%d] Adjusted inoffset to %d (after illegal attribute).",
		 tct->tid, tct->inoffset);
	return(tct->inbuf+tct->inoffset);
	}

  /* try to find value if we are not at the end of a tag*/
  if (attr_end[0]!='>')
    val_start=find_valstart(attr_end,tct);

  if (val_start==NULL) {
    /* no value */
    replaced=attr_end[0]; /* store char */
    attr_end[0]=0; /* terminate attribute string */

    process_attr(attr_start, NULL, tct);	
    attr_end[0]=replaced; /* put char back */
      /* adjust inoffset: skip possible WS between attrend and '>' */
    tct->inoffset=skip_leading_whitespace(attr_end)-tct->inbuf+1;
    logmsg(LOG_DEBUG, "[%d] Adjusted inoffset to %d (after attr).",
	   tct->tid, tct->inoffset);

    return attr_end;
    
  } else if (val_start==(char *)FAILURE) {
    /* illegal value. FIXME: adjust inoffset */
    logmsg(LOG_NOTICE, "[%d] Value for attribute '%s' contains illegal chars.",
	   tct->tid, attr_start);
    return  skip_to_next_whitespace(attr_end);
  }
  
  val_end=find_valend(val_start, tct);
  if (val_end==NULL) {
	logmsg(LOG_NOTICE, "[%d] Value for attribute '%s' contains illegal chars.",
	       tct->tid, attr_start);
	/* adjust inoffset: skip to next whitespace */
	return skip_to_next_whitespace(val_start);
  }
  
  replaced=val_end[0]; /* remember the char we're about to replace */
  attr_end[0]=0; val_end[0]=0;    /* terminate attribute and value */
  logmsg(LOG_DEBUG, "[%d] Processing attribute %s with value %s", tct->tid, attr_start, val_start);
  process_attr(attr_start, val_start,tct);     /* process it */
  /* put it back */
  val_end[0]=replaced;

  /* adjust inoffset: skip possible WS between valend and '>' */
  tct->inoffset=skip_leading_whitespace(val_end)-tct->inbuf+1;
  logmsg(LOG_DEBUG, "[%d] Adjusted inoffset to %d (after attr value).",
		 tct->tid, tct->inoffset);

  return val_end;
}

/*
 * do_tag
 *
 * Try to get and handle a tag in string. This is basically the same as
 * do_attr() except that a tag is slightly more than an attribute
 * (characterwise) -- the slash ('/') is a legal character, too.
 * See do_attr() for a description and notes.
 * Returns:
 *  PARSE_OKAY -- everything's normal, continue with attributes.
 *  PARSE_ERROR -- an error occurred, e.g. there were illegal charaters
 *  SKIP_TAG -- skip the tag completely.
 */
static int do_tag(char *buf, TCT *tct) {
  char *tag_start, *tag_end;
  int tag_result; char replaced;
    
  /* find tag */
  tag_start = find_tagstart(buf,tct);
  if (tag_start==NULL) {
	logmsg(LOG_NOTICE, "[%d] Tag '%s' starts with illegal characters.",
	       tct->tid, buf);
	return PARSE_ERROR;
  }
  
  /* find end of tag */
  tag_end = find_tagend(tag_start+1);
  /* illegal ? */
  if (tag_end==NULL) {
	logmsg(LOG_NOTICE, "[%d] Tag '%s' contains illegal characters.",
	       tct->tid, buf);
	return PARSE_ERROR;
  }
  
  /* terminate tag. remember the char we replace */
  replaced=tag_end[0];
  tag_end[0]=0;
  tag_result=process_tag(tag_start, tct);
  /* put back what we replaced. this is important for the caller. */
  tag_end[0]=replaced;
  
  /* adjust inoffset */
  tct->inoffset=tag_end-tct->inbuf;
  logmsg(LOG_DEBUG, "[%d] Adjusted inoffset to %d (tag end).",
		 tct->tid, tct->inoffset);

  if (tag_result==-1)
	/* abort: tag not allowed */
 	return SKIP_TAG;
  else
	/* okay, continue with attributes */
    return PARSE_OKAY;
}

/*
 * check_tag
 *
 * Performs a very rough check of the "tag":
 *   - !DOCTYPE... is copied
 *   - others are left untouched.
 *  MAYBE: add !ENTITY? This is SGML after all...
 */
static int check_tag(char *buf, TCT *tct) {
    
    /* is a doctype spec.? */
    if (strncmp(buf, "!DOCTYPE",8)==0) {
	/* yes, it is: just copy it to output buffer */
	  copy_doctype(buf,tct);
	  return DOCTYPE_LIKE;
    }

    /* otherwise expect to have a normal tag */
    return NORMAL_TAG;
}

static int copy_doctype(char *buf, TCT *tct) {
  char *s;

  s = skip_until_end_of_tag(buf);
  /*
   * CHECK if s[0] == 0
   * Since DOCTYPEs are at the beginning of a document
   * this would imply a biiig doctype. About HTTP_BUFSIZE.
   * By now we just copy it and set the offset. This seems to
   * suffice.
   */ 

  /* copy it */
  output_string("<", 1, tct);
  output_string(buf, s-buf, tct);

  /* adjust inoffset */
  tct->inoffset = s-tct->inbuf+1;
  return OKAY;
}

/*
 * parse_tag
 *
 * Parse the tag in the buffer.
 * Process the tag itself (do_tag()), check for attributes and
 * process them one by one if there are any (do_attr()).
 * Returns -1 on error, 0 otherwise.
 */
int parse_tag(char *buf, TCT *tct) {
  char *s=buf;
  int result;

  /* check for normal tag */
  if (check_tag(buf, tct)==DOCTYPE_LIKE) return OKAY; 
    
  result = do_tag(s,tct);
  logmsg(LOG_DEBUG, "[%d] Done with tag, proceeding with attrs.",
		 tct->tid);
    
  /* check for error */
  if (result==PARSE_ERROR) {
        /* error whilst parsing, skip this tag */
	output_string("<!-- malformed tag skipped -->", 30, tct);

	s = skip_until_end_of_tag(s);
	tct->inoffset=s-tct->inbuf;

	return OKAY; /* we handled this error */
  } else if (result==SKIP_TAG) {
	/* skip */
	output_string("<!-- unknown tag skipped -->", 28, tct);

	s = skip_until_end_of_tag(s);
	tct->inoffset=s-tct->inbuf;

	/* FIXME: skip this tag completely. Set inoffset... */
	return OKAY; /* CHECK: this return wasn't here before... why? */
  }
  
  s = tct->inbuf+tct->inoffset;
 
  /* check if there are any attrs to process */
  if (s[0]==0 || s[0]=='>') {
	/* no */
	logmsg(LOG_DEBUG, "[%d] This tag has no attributes.", tct->tid);
 	output_string(">", 1, tct); /* 'close' the tag */

    /* adjust inoffset */
    tct->inoffset=s-tct->inbuf+1;
	logmsg(LOG_DEBUG, "[%d] Adjusted inoffset to %d.",
		   tct->tid, tct->inoffset);

	return OKAY; /* and leave */
  } else {
	/* we know there are attributes */
	logmsg(LOG_DEBUG, "[%d] Processing attributes.", tct->tid);
 	while(s!= NULL && s!=(char *)-1) { /* s>0 doesn't work */ 
	  s=do_attr(s,tct);
	}
	logmsg(LOG_DEBUG, "[%d] Done with attributes.", tct->tid);
	output_string(">", 1, tct);
  }
  
  return OKAY;
}


/*
 * process_tag
 *
 * Takes buffer containing the tag and the context.
 * Look up the tag in hash-table built from the configuration and
 * act like the action says.
 * Input: terminated tag and thread context
 * Returns:
 *  0  -- continue with the attributes
 * -1  -- tag not allowed, skip rest of tag string
 */
static int process_tag(char *tagbuf, TCT *tct) {
    char *replacement; /* from hash-table */
    int action; /* from hash-table */

	logmsg(LOG_DEBUG, "[%d] Processing tag %s.",
		   tct->tid, tagbuf);
    
    /* find out what we are supposed to do */
    action = tag_action(tagbuf);
    
    if (action < 0) {
	/* not specified in configuration file */
 	logmsg(LOG_NOTICE, "[%d] Not allowing tag '%s'.", tct->tid, tagbuf);
	return -1;
    }
    
    if (action & (ACT_REPLABORT | ACT_REPLCONT | ACT_REPLSKIP |
                  ACT_REPLENDSKIP)) {
	/* get replacement for this tag*/
	replacement=tag_replacement(tagbuf);
	if (action & ACT_LOG) {
	    logmsg(LOG_NOTICE, "[%d] LOG: Replacing %s with '%s'.",
		   tct->tid, tagbuf, replacement);
	}
	
	output_string("<", 1, tct);
	output_string(replacement, strlen(replacement),tct);
		
	if (action & ACT_REPLSKIP) {
	    output_string(">", 1, tct);
	    tct->skip_flag=1; /* skip the following contents */
	    return -1; /* don't proceed with the attributes */
	} else if (action & ACT_REPLENDSKIP) {
	    /* turn off skipping mode and write the replacement */
	    tct->skip_flag=0;
	    output_string("<", 1, tct);
	    output_string(replacement, strlen(replacement),tct);
	    output_string(">", 1, tct);
	    
	    return -1; /* don't proceed with the attributes */
	} else if (action & ACT_REPLABORT) {
	    output_string(">", 1, tct);
	    return -1; /* don't proceed with the attributes */
	} else {
	    return 0; /* proceed with the attributes */
	}
    } else if (action & ACT_ALLOW) {
	if (action & ACT_LOG) {
	    logmsg(LOG_NOTICE, "[%d] LOG: Allowing tag %s.", tct->tid, tagbuf);
	}
	
	/* copy tag 'as is' to output buffer */
	output_string("<", 1, tct);
	output_string(tagbuf, strlen(tagbuf), tct);
	return 0; /* proceed with the attributes */
    } else if (action & ACT_LOG) {
	/* just log it */
	logmsg(LOG_NOTICE, "[%d] LOG: Skipping tag: %s", tct->tid, tagbuf);
	return -1; /* don't proceed with the attributes */
    }
    
    logmsg(LOG_ERR, "[%d] Action %d not processed", tct->tid, action);
    logmsg(LOG_ERR, "[%d] This is a bug: FIXME", tct->tid);
    return -1;
}


/*
 * process_attr
 *
 */
static char *process_attr(char*attrbuf, char *valbuf, TCT *tct) 
{
    char *replacement; /* from hash-table */
    int action; /* from hash-table */
    
    /* find out what we are supposed to do */
    action = attrib_action(attrbuf);
    
    if (action < 0) {
	/* action not specified in configuration file */
 	logmsg(LOG_NOTICE, "[%d] Not allowing attribute '%s'.", tct->tid, attrbuf);
	return (char *)-1;
    }

    /* check if we have some script code in the value */
    if (check_for_script(valbuf)!=0){
	logmsg(LOG_NOTICE, "Found script code in value of attribute %s:", attrbuf);
	logmsg(LOG_NOTICE, "\"%s\"", valbuf);
	return (char *)-1;
    }

    /* logging? */
    {
	int level;
	if (action & ACT_LOG)
	    level=LOG_NOTICE;
	else
	    level=LOG_DEBUG;
	
	if (valbuf!=NULL) {
	    logmsg(level, "[%d] Processing attr '%s' with value '%s'.",
		   tct->tid, attrbuf, valbuf);
	} else
	    logmsg(level, "[%d] Processing attr '%s'.",
		   tct->tid, attrbuf);
    }


    /* you can either replace an attr or allow it */

    if (action & (ACT_REPLABORT | ACT_REPLCONT)) {
	  /* get replacement for this attr*/
	  replacement=attrib_replacement(attrbuf);
	  if (action & ACT_LOG) {
	      logmsg(LOG_NOTICE, "[%d] Replacing with '%s'.",
		     tct->tid, replacement);
	  }
	  
	  output_string(" ", 1, tct);
	  output_string(replacement, strlen(replacement),tct);
	  if (action & ACT_REPLABORT) {
	    return (char *)-1;
	  } else {
	    return 0; /* proceed with the attributes */
	  }
    } else if (action & ACT_ALLOW) {
       if (unallowed_href(attrbuf, valbuf)) {
          if (valbuf == NULL) {
              logmsg(LOG_NOTICE, "[%d] Unallowed HREF: no attribute", tct->tid);
	  } else {
              logmsg(LOG_NOTICE, "[%d] Unallowed HREF: '%s'", tct->tid, valbuf);
          }
       } else {
	  if (action & ACT_LOG) {
	      logmsg(LOG_NOTICE, "[%d] Allowing attribute.", tct->tid);
	  }
	  
	  /* copy attr 'as is' to output buffer */
	  output_string(" ", 1, tct);
	  output_string(attrbuf, strlen(attrbuf), tct);
	  if (valbuf!=NULL) {
	      /* we have a value */
	      output_string("=", 1, tct);
	      output_string(valbuf, strlen(valbuf), tct);
	  }
       }
       return 0; /* proceed with the attributes */
    } else if (action & ACT_LOG) {
	/* just log it */
	/*logmsg(LOG_NOTICE, " (no further action)"); */
	return 0; /* proceed with the attributes */
    }
    
    logmsg(LOG_ERR, "[%d] Action %d not processed.", tct->tid, action);
    logmsg(LOG_ERR, "[%d] This is a bug: FIXME.", tct->tid);
    return (char *)-1;
    
	    
}

/*
 * check_for_script
 *
 * Checks the value buffer for embedded script code.
 * (This script code is introduced by "&{".)
 * Returns 0 if the buffer does not contain "&{",
 * -1 otherwise.
 * Notes: Safe to call with a value of NULL for the buffer.
 */
static int check_for_script(char *valbuf) 
{
    char *pos;

    /* return ok if we have no value */
    if (valbuf==NULL)
	return 0;

    pos = valbuf;

    /* look for ampersand */
    while ((pos = strnchr(pos, '&', strlen(pos))) != NULL) {
      if (strlen(pos)>1) {
	if (pos[1]=='{') return -1;	
      }
      pos++;
    }

    /* no script code found */
    return 0;
}

/*
 * recover_from_parse_error()
 *
 * This will be called when an error occurs whilst
 * parsing an HTML document. It sets the INOFFSET to
 * the last valid point. This can be right after the
 * last tag, attribute or value (whatever is highest).
 */
int recover_from_parse_error(TCT *tct) {
  char *whitespace;

  logmsg(LOG_DEBUG,
		 "[%d] Trying to recover from error whilst parsing.",
		 tct->tid);

  logmsg(LOG_DEBUG,
		 "[%d] Pick highest of %d,%d,%d and skip to whitespace.",
		 tct->tid,
		 tct->last_valid_value-tct->inbuf,
		 tct->last_valid_attr-tct->inbuf,
		 tct->last_valid_tag-tct->inbuf);
 
  if (tct->last_valid_value>tct->last_valid_attr) /* val>attr? */
	if (tct->last_valid_value>tct->last_valid_tag) /* val>tag? */
	  whitespace = skip_to_next_whitespace(tct->last_valid_value); /* val>tag! */
	else  /* tag > val */
	  whitespace = skip_to_next_whitespace(tct->last_valid_tag);
  else /* attr > val! */
	if (tct->last_valid_attr>tct->last_valid_tag) /* attr>tag? */
	  whitespace = skip_to_next_whitespace(tct->last_valid_attr); /* attr>tag! */
	else  /* tag > attr */
	  whitespace = skip_to_next_whitespace(tct->last_valid_tag);
	
  /* there should be some whitespace */
  if (whitespace == NULL)
	return FAILURE;
  else
	logmsg(LOG_DEBUG,
		   "[%d] Whitespace found at %d", tct->tid, whitespace-tct->inbuf);
  
  tct->inoffset = whitespace-tct->inbuf;
  return OKAY;
}

/*
 * skip_until_end_of_tag()
 *
 * returns address of first '>' after BUF or end of string.
 */
char *skip_until_end_of_tag(char *buf) {
  int count_angle=1, count_quote=0, i=0;

  /* step over tag FIXME: this is quick hack */
  while (buf[i] && !(count_angle==0 && (count_quote%2==0))) {
	if (buf[i]=='\"') count_quote++;
	if (buf[i]=='<') count_angle++;
	if (buf[i]=='>') count_angle--;
	i++;
  }
  return buf+i;
}
