/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 *
 * $Id: read.c,v 1.12 2000/05/07 16:18:11 glauschwuffel Exp $
 *
 * $Log: read.c,v $
 * Revision 1.12  2000/05/07 16:18:11  glauschwuffel
 * Input data is now terminated with a null character, therefore we read only HTTP_BUFZIE-1 bytes.
 *
 * Revision 1.11  2000/04/24 19:52:04  gellert
 * BUGFIX: if we had an incomplete tag in the buffer, we did a
 * recv and assumed, that the buffer then is full. this is not
 * true if the server has a small segment size and we did just
 * get one segment til the recv. than the tag may still be
 * incomplete and we throw it away. now we recv until the buffer
 * is full (or eof occured) explicitly.
 *
 * Revision 1.10  2000/03/07 15:18:01  gellert
 * some changes to port to linux!!!
 *
 * Revision 1.9  2000/03/07 15:06:56  goldbach
 * added GPL info
 *
 * Revision 1.8  2000/03/01 17:43:56  goldbach
 * introduced some constants for read and copy-pkts
 *
 * Revision 1.7  1999/11/30 16:57:53  gellert
 * skip flag: initialization moved to beginning of copy-pakets. read_header: find_empty_line is called with the length of the read block, not the full HTTP_BUFSIZE.
 *
 * Revision 1.6  1999/11/11 15:59:33  gellert
 * If there is no empty line in the first blockof the header,
 * a new read is done to get the rest of the header (until the
 * inbuf is filled).
 *
 * Revision 1.5  1999/09/21 09:56:28  gellert
 * just changed the log level of the "EOF" messages to debug.
 *
 * Revision 1.4  1999/07/28 14:23:40  gellert
 * some changes to the calls of logmsg
 *
 * Revision 1.3  1999/05/27 16:48:01  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.2  1999/05/12 16:04:23  goldbach
 * major clean-up session
 *
 * Revision 1.1  1999/05/12 14:39:36  goldbach
 * Renamed read_header.c to read.c
 *
 * Revision 1.6  1999/05/05 11:58:52  goldbach
 * minor clean-ups
 *
 * Revision 1.5  1999/03/30 12:30:08  goldbach
 * 1st surfable version
 *
 * Revision 1.4  1999/03/02 11:42:13  goldbach
 * sets inlen: the inbuf is valid up to here.
 *
 * Revision 1.3  1999/02/18 15:10:27  goldbach
 * fixed return codes, cleaned up the source a bit
 *
 * Revision 1.2  1999/02/16 17:42:58  gellert
 * fixed the missing end of comment "* /" before the call to strstr.
 * replaced "strstr" with an own routine to search for the end of the header.
 * this new routine is given a maximum length to search.
 *
 * Revision 1.1  1999/02/15 16:14:07  goldbach
 * Erste Version
 *
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <syslog.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "httpf.h"
#include "mythreads.h"
#include "read.h"
#include "misc.h"

static const char nls[4]={0x0d,0x0a,0x0d,0x0a};

static int find_empty_line_in_header(TCT *, int);
static int read_block(TCT *, int);
static int read_header(TCT *, int);


/*
 * read_block
 *
 * Read data block from socket into inbuf. The address to read data into
 * is inbuf+inlen. The buffer is filled up to the end.
 *
 * To read as much data as possible do:
 *   - reset_buffers()
 *   - read_blcok(_)
 * To fill the inbuf from a specific position:
 *   - set inlen to the position
 *   - read_block(_) [tries to read HTTP_BUFZIE-tct->inlen bytes]
 *
 * Sets:
 *   - tct->inbuf contents with data read from tct->insocket
 *   - tct->inlen to new length of valid data in inbuf.
 *   
 * Returns:
 *   1 -- read error on socket 
 *   2 -- eof
 *   0 -- inbuf contains valid data.
 *
 */
static int read_block(TCT *tct, int socket)
{
    int len;
    
    /* read block */
    len=recv(socket,
	     tct->inbuf+tct->inlen,
	     HTTP_BUFSIZE-tct->inlen-1, 0);

    /* error on socket */
    if (len < 0) {
	logmsg(LOG_ERR, "[%d] Read error on socket %d with errno %d.",
	       tct->tid, socket, errno);
	return READ_SOCKET_ERROR;
    }

    /* eof */
    if (len==0) {
	logmsg(LOG_DEBUG, "[%d] EOF on socket %d.",
	       tct->tid, socket);
       	return READ_EOF;
    }

    /* add #of bytes read to valid buffer length */
    if(len>=0) tct->inlen += len;

	/* terminate data with null character */
    tct->inbuf[tct->inlen]=0;

    logmsg(LOG_DEBUG, "[%d] Terminated data after '%c%c%c%c'",
	   tct->tid, tct->inbuf[tct->inlen-4], tct->inbuf[tct->inlen-3],
		   tct->inbuf[tct->inlen-2], tct->inbuf[tct->inlen-1]);

    logmsg(LOG_DEBUG, "[%d] Read %d bytes from socket %d.",
	   tct->tid, len);
    logmsg(LOG_DEBUG, "[%d] Inbuf contains %d bytes.",
	   tct->tid, tct->inlen);

    return READ_OK;
}


/*
 * read_header
 *
 * Read a single buffer of HTTP_BUFSIZE bytes via recv().
 * Try to find the end of the HTTP header and set TCT->inoffset
 * to the offset where the body begins.
 *
 * Results:
 *   0    no error
 *   1    error on insocket (from read_block())
 *   2    eof (from read_block())
 *   3    end of header not found
 */
static int read_header(TCT *tct, int socket)
{
    int nl, result;
    int quit = 0;
    
    tct->inoffset = 0;
    tct->inlen = 0;

    result = read_block(tct, socket);
    if (result!=READ_OK) return result;
    
   /*
    * Try to find the end of the header (an empty line).
    * Don't use strstr() since our buffer doesn't have
    * a trailing null character.
    */

    do {

      nl = find_empty_line_in_header(tct, tct->inlen);

      if (nl==(int) NULL) {
	if (tct->inlen<HTTP_BUFSIZE) {
	  logmsg(LOG_DEBUG, "[%d] Reading next block of header!", tct->tid);
	  result = read_block(tct, socket);
	  if (result!=READ_OK) return result;
	} else {
	  logmsg(LOG_ERR, "[%d] No end of HTTP header found.",
	       tct->tid);
	  return READ_NO_END_OF_HEADER;
	}
      } else {
        quit=1;
      }
    } while (quit==0);
    
    /* the line was found and we can now tell where the body begins */
    tct->inoffset = (int)nl;
    
    logmsg(LOG_DEBUG, "[%d] Body starts at offset %d.",
	   tct->tid, tct->inoffset);
        
    return READ_OK;
}

static int find_empty_line_in_header(TCT *tct, int length) {

    int n,i = 0;
    char *buffend;
    
    buffend = tct->inbuf+length;
    for (n=0; n<length; n++) {
       /* is this an expected char? */
       if (*(tct->inbuf+n) == nls[i]) {
	  i++;
	  if (i == sizeof(nls)) {
	      logmsg(LOG_DEBUG, "[%d] Header ends at offset %d.",
		     tct->tid, n+1);
	      
	      return (n+1); }
       /* not expected, so begin to search again for the first
          char of the searchstring */
       } else {
          i=0;
       }
    }
    return ((int) NULL);
}

int read_client_header(TCT *tct)
{
    logmsg(LOG_INFO, "[%d] Read client header.",
	   tct->tid, tct->inlen);
    return read_header(tct, tct->insocket);
}

int read_server_header(TCT *tct)
{
    logmsg(LOG_INFO, "[%d] Read server header.",
	   tct->tid, tct->inlen);
    return read_header(tct, tct->outsocket);
}

int read_client_body(TCT *tct)
{
    logmsg(LOG_INFO, "[%d] Read client body.",
	   tct->tid, tct->inlen);
    return read_block(tct, tct->insocket);
}

int read_server_body(TCT *tct)
{
    logmsg(LOG_INFO, "[%d] Read server body.",
	   tct->tid,  tct->inlen);
    return read_block(tct, tct->outsocket);
}

/*
 * read_full_server_block()
 *
 * Note: HTTP_BUFSIZE-1 is the maximum size since we set the last
 * char in the buffer to null (termination).
 */
int read_full_server_block(TCT *tct)
{
	int result = READ_OK;

	while ((tct->inlen < HTTP_BUFSIZE-1) && (result == READ_OK)) {
	  result = read_block(tct, tct->outsocket);
	}
    if (result==READ_EOF) tct->more_input_data=NO;

	return (result);
}
